/*
    ipodfs - a user-space filesystem for accessing iPods
    Copyright (C) 2006 Richard Dawe

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <ctype.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/vfs.h>
#include <sys/statvfs.h>

#define FUSE_USE_VERSION 25
#include <fuse.h>

#include <gpod/itdb.h>

static Itdb_iTunesDB *itunes = NULL;
static const char *ipod_path = NULL;
static struct statvfs ipod_statvfs;

/* XXX: Something better than O(N). */
static int
resolve (const char *path, char **resolved_path, Itdb_Track **resolved_track)
{
  *resolved_path = NULL;
  *resolved_track = NULL;

  if (strcmp(path, "/") == 0)
    {
      /* Root */
      *resolved_path = strdup(ipod_path);
    }
  else
    {
      const int n_tracks = itdb_tracks_number(itunes);
      int i;

      /* Track */
      ++path;

      for (i = 0; i < n_tracks; ++i)
	{
	  Itdb_Track *track = g_list_nth_data(itunes->tracks, i);
	  char *track_path = itdb_filename_on_ipod(track);
	  char *filename = NULL;

	  filename = strrchr(track_path, '/');
	  if (filename)
	    ++filename;
	  else
	    filename = track_path;

	  if (strcmp(filename, path) == 0)
	    {
	      *resolved_path = track_path;
	      *resolved_track = track;
	      break;
	    }

	  free(track_path);
	}
    }

  return *resolved_path && *resolved_track;
}

static int
write_dbs (Itdb_iTunesDB *db)
{
  GError *err = NULL;
  int ok = 1;

  itdb_write(db, &err);
  if (err != NULL)
    {
      fprintf (stderr, "Unable to write iTunes database: %s\n", err->message);
      g_error_free (err);
      err = NULL;
      ok = 0;
    }

  itdb_shuffle_write(db, &err);
  if (err != NULL)
    {
      fprintf (stderr, "Unable to write iTunes shuffle database: %s\n", err->message);
      g_error_free (err);
      err = NULL;
      ok = 0;
    }

  return ok;
}

static int
ipodfs_getattr (const char *path, struct stat *sbuf)
{
  int cur_errno = ENOENT;
  int found = 0;

  if (strcmp(path, "/") == 0)
    {
      struct stat root_sbuf;

      memset(&root_sbuf, 0, sizeof(root_sbuf));
      if (stat(ipod_path, &root_sbuf) == -1)
	{
	  cur_errno = errno;
	}
      else
	{
	  sbuf->st_dev = root_sbuf.st_dev;
	  sbuf->st_ino = 0;
	  sbuf->st_mode = S_IFDIR | S_IRWXU;
	  sbuf->st_nlink = 1;
	  sbuf->st_uid = getuid();
	  sbuf->st_gid = getgid();
	  sbuf->st_size = 0;
	  sbuf->st_blksize = ipod_statvfs.f_bsize;
	  sbuf->st_blocks = 1;
	  sbuf->st_atime = root_sbuf.st_atime;
	  sbuf->st_mtime = root_sbuf.st_mtime;
	  sbuf->st_ctime = root_sbuf.st_ctime;

	  found = 1;
	}
    }
  else if (*path == '/')
    {
      /* Track */
      char *track_path = NULL;
      Itdb_Track *track = NULL;

      if (resolve(path, &track_path, &track))
	{
	  struct stat track_sbuf;
	  int ret;

	  memset(&track_sbuf, 0, sizeof(track_sbuf));
	  ret = stat(track_path, &track_sbuf);
	  if (ret == -1)
	    {
	      cur_errno = errno;
	    }
	  else
	    {
	      sbuf->st_dev = track_sbuf.st_dev;
	      sbuf->st_ino = track->id;
	      sbuf->st_mode = S_IFREG | S_IRUSR | S_IWUSR;
	      sbuf->st_nlink = 1;
	      sbuf->st_uid = getuid();
	      sbuf->st_gid = getgid();
	      sbuf->st_size = track_sbuf.st_size;
	      sbuf->st_blksize = ipod_statvfs.f_bsize;
	      sbuf->st_blocks
		= (track->size + ipod_statvfs.f_bsize)
		/ ipod_statvfs.f_bsize;
	      sbuf->st_atime = track_sbuf.st_atime;
	      sbuf->st_mtime = track_sbuf.st_mtime;
	      sbuf->st_ctime = track_sbuf.st_ctime;

	      found = 1;
	    }
	}

      if (track_path)
	free(track_path);
    }

  return found ? 0 : -cur_errno;
}

static int
ipodfs_mknod (const char *path, mode_t mode, dev_t rdev)
{
  int cur_errno = ENOENT;
  int found = 0;
  int ret = 0;

  if (strcmp(path, "/") == 0)
    {
      /* Root */
      cur_errno = EISDIR;
    }
  else
    {
      /* Track */

      /* XXX: Guess the file type */
      /* XXX: Mark the file as new, so we add it to the DB on release */
      if (!S_ISREG(mode) || (strchr(path + 1, '/') != NULL))
	{
	  /*
	   * Only allow regular files. Only allow files in the current
	   * directory.
	   */
	  cur_errno = EPERM;
	}
      else
	{
	  const int n_dirs = itdb_musicdirs_number(itunes);
	  char track_path[PATH_MAX];
	  char track_ipod_path[PATH_MAX];
	  Itdb_Track *track = NULL;
	  int n;

	  ++path; /* Skip leading '/' */

	  /* Generate a filename in one of the Fxx directories */
	  n = rand() % n_dirs;
	  snprintf(track_path, sizeof(track_path),
		   "%s/iPod_Control/Music/F%02d/%s", ipod_path, n, path);
	  snprintf(track_ipod_path, sizeof(track_ipod_path),
		   "iPod_Control:Music:F%02d:%s", n, path);

	  /* XXX: Add to the master playlist. */
	  track = itdb_track_new();
	  if (!track)
	    {
	      cur_errno = ENOMEM;
	    }
	  else
	    {
	      char *ext = NULL;

	      /*
	       * Use the file extension as the description. Crap,
	       * but this is the best way of getting the filetype
	       * set up at the start. We can do better once some data
	       * has been read to the file - e.g.: using ID3 tags.
	       */
	      ext = strrchr(path, '.');
	      if (ext)
		{
		  ++ext;
		  ext = strdup(ext);
		}
	      if (ext)
		{
		  char *p;
		  for (p = ext; *p; p++) { *p = toupper(*p); }
		}

	      track->ipod_path = strdup(track_ipod_path);
	      track->description = ext;

	      ret = mknod(track_path, mode, rdev);
	      if (ret == -1)
		{
		  cur_errno = errno;
		}
	      else
		{
		  Itdb_Playlist *master_pl = itdb_playlist_mpl(itunes);

		  itdb_track_add(itunes, track, -1);
		  itdb_playlist_add_track(master_pl, track, -1);

		  /* Write out the DBs */
		  if (write_dbs(itunes))
		    found = 1;
		}
	    }
	}
    }

  return found ? 0 : -cur_errno;
}

static int
ipodfs_unlink (const char *path)
{
  int cur_errno = ENOENT;
  int found = 0;

  if (strcmp(path, "/") == 0)
    {
      /* Root */
      cur_errno = EISDIR;
    }
  else
    {
      /* Track */
      char *track_path = NULL;
      Itdb_Track *track = NULL;

      if (resolve(path, &track_path, &track))
	{
	  const int n_playlists = itdb_playlists_number(itunes);
	  int i;

	  /* Remove the track from any playlists */
	  for (i = 0; i < n_playlists; i++)
	    {
	      Itdb_Playlist *pl = g_list_nth_data(itunes->playlists, i);

	      itdb_playlist_remove_track(pl, track);
	    }

	  /* Remove the track from the DB */
	  itdb_track_remove(track);

	  /* Write out the DBs */
	  if (write_dbs(itunes))
	    {
	      /* Finally physically remove the file */
	      int ret = unlink(track_path);
	      if (ret == -1)
		cur_errno = errno;
	      else
		found = 1;
	    }
	}

      if (track_path)
	free(track_path);
    }

  return found ? 0 : -cur_errno;
}

static int
replace_last_component (const char *str, const char sep, const char *replacement, char *buf, const size_t buflen)
{
  char *p = NULL;

  strncpy(buf, str, buflen);
  buf[buflen - 1] = '\0';

  p = strrchr(buf, sep);
  if (p)
    {
      /*
       * Path separator - replace last path component
       * with new filename.
       */
      size_t spc;

      ++p;
      spc = buflen - (ptrdiff_t) (p - buf);
      strncpy(p, replacement, spc);
      p[spc - 1] = '\0';
    }
  else
    {
      /* No path separator -- weird. Just use new filename. */
      strncpy(buf, replacement, buflen);
    }

  /* XXX: Detect truncation */
  return 0;
}

static int
ipodfs_rename (const char *oldpath, const char *newpath)
{
  int cur_errno = ENOENT;
  int found = 0;

  if (strcmp(oldpath, "/") == 0)
    {
      /* Root */
      cur_errno = EPERM;
    }
  else
    {
      /* Track */
      char *track_oldpath = NULL;
      Itdb_Track *track = NULL;

      if (resolve(oldpath, &track_oldpath, &track))
	{
	  char track_newpath[PATH_MAX];
	  char track_ipod_newpath[PATH_MAX];

	  /* Skip over root directory. */
	  ++newpath;

	  /* Build new names. */
	  replace_last_component(track_oldpath, '/', newpath,
				 track_newpath, sizeof(track_newpath));

	  replace_last_component(track->ipod_path, ':', newpath,
				 track_ipod_newpath, sizeof(track_ipod_newpath));

	  /* Rename the file */
	  if (rename(track_oldpath, track_newpath) == 0)
	    {
	      /* Update the playlist entry. */
	      free(track->ipod_path);
	      track->ipod_path = strdup(track_ipod_newpath);
	      if (!track->ipod_path)
		cur_errno = ENOMEM;

	      /* Write out the DBs */
	      if (track->ipod_path && write_dbs(itunes))
		found = 1;
	    }
	  else
	    {
	      cur_errno = errno;
	    }
	}

      if (track_oldpath)
	free(track_oldpath);
    }

  return found ? 0 : -cur_errno;
}

static int
ipodfs_truncate (const char *path, off_t size)
{
  int cur_errno = ENOENT;
  int found = 0;

  if (strcmp(path, "/") == 0)
    {
      /* Root */
      cur_errno = EISDIR;
    }
  else
    {
      /* Track */
      char *track_path = NULL;
      Itdb_Track *track = NULL;

      if (resolve(path, &track_path, &track))
	{
	  int ret = truncate(track_path, size);
	  if (ret == -1)
	    cur_errno = errno;
	  else
	    found = 1;
	}

      if (track_path)
	free(track_path);
    }

  return found ? 0 : -cur_errno;
}

static int
ipodfs_utime (const char *path, struct utimbuf *buf)
{
  int cur_errno = ENOENT;
  int found = 0;
  char *track_path = NULL;
  Itdb_Track *track = NULL;

  if ((strcmp(path, "/") == 0) || resolve(path, &track_path, &track))
    {
      if (utime(track_path, buf) == -1)
	cur_errno = errno;
      else
	found = 1;
    }

  return found ? 0 : -cur_errno;
}

static int
ipodfs_open (const char *path, struct fuse_file_info *fi)
{
  int cur_errno = ENOENT;
  int found = 0;

  if (strcmp(path, "/") == 0)
    {
      /* Root */
      cur_errno = EISDIR;
    }
  else
    {
      /* Track */
      char *track_path = NULL;
      Itdb_Track *track = NULL;

      if (resolve(path, &track_path, &track))
	{
	  /* Existing track */
	  /* XXX: What should we do if the file is already open? */
	  int fd = open(track_path, fi->flags);
	  if (fd >= 0)
	    {
	      fi->fh = fd; /* XXX: fd == 0 => need to dup to get non-zero? */

	      found = 1;
	    }
	  else
	    {
	      fi->fh = 0;

	      cur_errno = errno;
	    }
	}

      if (track_path)
	free(track_path);
    }

  return found ? 0 : -cur_errno;
}

static int
ipodfs_read (const char *path, char *buf, size_t size, off_t offset,
	     struct fuse_file_info *fi)
{
  int cur_errno = EBADF;
  int found = 0;
  int ret = 0;

  if (fi->fh)
    {
      ret = lseek(fi->fh, offset, SEEK_SET);
      if (ret == -1)
	{
	  cur_errno = errno;
	}
      else
	{
	  ret = read(fi->fh, buf, size);
	  if (ret == -1)
	    cur_errno = errno;
	  else
	    found = 1;
	}
    }

  return found ? ret : -cur_errno;
}

static int
ipodfs_write (const char *path, const char *buf, size_t size,
	      off_t offset, struct fuse_file_info *fi)
{
  int cur_errno = EBADF;
  int found = 0;
  int ret = 0;

  if (fi->fh)
    {
      ret = lseek(fi->fh, offset, SEEK_SET);
      if (ret == -1)
	{
	  cur_errno = errno;
	}
      else
	{
	  ret = write(fi->fh, buf, size);
	  if (ret == -1)
	    cur_errno = errno;
	  else
	    found = 1;
	}
    }

  return found ? ret : -cur_errno;
}

static int
ipodfs_statfs (const char *path, struct statvfs *sbuf)
{
  int ret;

  ret = statvfs(ipod_path, sbuf);
  if (ret == -1)
    ret = -errno;

  return ret;
}

static int
ipodfs_release (const char *path, struct fuse_file_info *fi)
{
  int cur_errno = ENOENT;
  int found = 0;

  if (fi->fh)
    {
      close(fi->fh);
      fi->fh = 0;

      found = 1;
    }
  /* XXX: Something needed for other cases? */

  return found ? 0 : -cur_errno;
}

static int
ipodfs_fsync (const char *path, int isdatasync, struct fuse_file_info *fi)
{
  int cur_errno = EBADF;
  int found = 0;

  if (fi->fh)
    {
      int ret = 0;

      if (isdatasync)
	ret = fdatasync(fi->fh);
      else
	ret = fsync(fi->fh);

      if (ret == -1)
	cur_errno = errno;
      else
	found = 1;
    }
  /* XXX: Something needed for other cases? */

  return found ? 0 : -cur_errno;  
}

static int
ipodfs_readdir (const char *path,
		void *buf,
		fuse_fill_dir_t filler,
		off_t offset,
		struct fuse_file_info *fi)
{
  const int n_tracks = itdb_tracks_number(itunes);
  int found = 0;
  int i;

  if (strcmp(path, "/") == 0)
    {
      for (i = 0; i < n_tracks; ++i)
	{
	  Itdb_Track *track = g_list_nth_data(itunes->tracks, i);
	  char *track_path = itdb_filename_on_ipod(track);
	  char *filename = NULL;
	  struct stat sbuf;

	  filename = strrchr(track_path, '/');
	  if (filename)
	    ++filename;
	  else
	    filename = track_path;
	  
	  memset(&sbuf, 0, sizeof(sbuf));
	  sbuf.st_ino = track->id;
	  sbuf.st_mode = S_IFREG;
	  if (filler(buf, filename, &sbuf, 0))
	    {
	      free(track_path);
	      break;
	    }

	  free(track_path);
	}

      found = 1;
    }

  return found ? 0 : -ENOENT;
}

static struct fuse_operations ipodfs_operations = {
  ipodfs_getattr,
  /* ipodfs_readlink */ NULL,
  /* ipodfs_getdir */ NULL,
  ipodfs_mknod,
  /* ipodfs_mkdir */ NULL,
  ipodfs_unlink,
  /* ipodfs_rmdir */ NULL,
  /* ipodfs_symlink */ NULL,
  ipodfs_rename,
  /* ipodfs_link */ NULL,
  /* ipodfs_chmod */ NULL,
  /* ipodfs_chown */ NULL,
  ipodfs_truncate,
  ipodfs_utime,
  ipodfs_open,
  ipodfs_read,
  ipodfs_write,
  ipodfs_statfs,
  /* ipodfs_flush */ NULL,
  ipodfs_release,
  ipodfs_fsync,
  /* ipodfs_setxattr */ NULL,
  /* ipodfs_getxattr */ NULL,
  /* ipodfs_listxattr */ NULL,
  /* ipodfs_removexattr */ NULL,
  /* ipodfs_opendir */ NULL,
  ipodfs_readdir,
  /* ipodfs_releasedir */ NULL,
  /* ipodfs_fsyncdir */ NULL,
  /* ipodfs_init */ NULL,
  /* ipodfs_destroy */ NULL,
  /* ipodfs_access */ NULL,
  /* ipodfs_create */ NULL,
  /* ipodfs_ftruncate */ NULL,
  /* ipodfs_fgetattr */ NULL,
};

static void
syntax (const char *program_name)
{
  fprintf(stderr, "Syntax: %s <iPod-path> <mount-point>\n", program_name);
}

int
main (int argc, char *argv[])
{
  const char *program_name = argv[0];
  const char *mount_point = NULL;
  const char *p = NULL;
  GError *err = NULL;

  p = strrchr(program_name, '/');
  if (p)
    {
      ++p;
      program_name = p;
    }

#if 0
  if (argc < 2 /*3*/)
    {
      syntax(program_name);
      return EXIT_FAILURE;
    }
#endif

  /* XXX: Parse args */
  mount_point = argv[1];
  ipod_path = "/media/IPOD"; /*argv[2]*/

  if (statvfs(ipod_path, &ipod_statvfs) == -1)
    {
      perror(program_name);
      return EXIT_FAILURE;
    }

  itunes = itdb_parse(ipod_path, &err);
  if (err != NULL)
    {
      fprintf (stderr, "Unable to open iTunes database: %s\n", err->message);
      g_error_free (err);
    }

  return
      fuse_main(argc, argv, &ipodfs_operations)
    ? EXIT_SUCCESS
    : EXIT_FAILURE;
}
